package main

import (
	"consumerDemo-avro_go/avro"
	"consumerDemo-avro_go/avro/gen"
	"consumerDemo-avro_go/sasl_sarama"
	"context"
	"encoding/hex"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"github.com/Shopify/sarama"
	"strconv"
	"strings"
	"time"
)

var (
	partitionConsumers = make(map[int32]*partitionMsgConsumer)
	brokerList         = flag.String("brokers", "", "REQUIRED: The comma separated list of brokers in the Kafka cluster")
	topic              = flag.String("topic", "", "REQUIRED: the topic to consume")
	group              = flag.String("group", "", "REQUIRED: the consumer group")
	user               = flag.String("user", "", "the user name")
	password           = flag.String("password", "", "the password")
	trans2sql          = flag.Bool("trans2sql", false, "trans message to sql or not")
)

func main() {
	flag.Parse()
	config := sarama.NewConfig()
	config.Version = sarama.V0_11_0_0
	config.Consumer.Return.Errors = true
	config.Consumer.Offsets.Initial = sarama.OffsetOldest
	config.Consumer.Fetch.Default = 20 << 20 // kafka消息上限20M
	if *user != "" {
		config.Net.SASL.Enable = true
		config.Net.SASL.User = *user
		config.Net.SASL.Password = *password
		config.Net.SASL.Handshake = true
		config.Net.SASL.SCRAMClientGeneratorFunc = func() sarama.SCRAMClient { return &sasl_sarama.XDGSCRAMClient{HashGeneratorFcn: sasl_sarama.SHA512} }
		config.Net.SASL.Mechanism = sarama.SASLTypeSCRAMSHA512
	}

	// Start with a client
	client, err := sarama.NewClient(strings.Split(*brokerList, ","), config)
	if err != nil {
		panic(err)
	}
	defer func() { _ = client.Close() }()

	// Start a new consumer group
	group, err := sarama.NewConsumerGroupFromClient(*group, client)
	if err != nil {
		panic(err)
	}
	defer func() { _ = group.Close() }()

	// Track errors
	go func() {
		for err := range group.Errors() {
			fmt.Println("ERROR", err)
		}
	}()

	// Iterate over consumer sessions.
	ctx := context.Background()
	for {
		handler := demoConsumerGroupHandler{}
		// 开始消费
		err := group.Consume(ctx, []string{*topic}, handler)
		if err != nil {
			panic(err)
		}
	}
}

type demoConsumerGroupHandler struct {
}

func (demoConsumerGroupHandler) Setup(sess sarama.ConsumerGroupSession) error {
	setup := make(map[int32]bool)
	for _, part := range sess.Claims()[*topic] {
		if _, ok := partitionConsumers[part]; !ok {
			partitionConsumers[part] = &partitionMsgConsumer{}
		}

		setup[part] = true
	}

	var invalid []int32
	for p := range partitionConsumers {
		if _, ok := setup[p]; !ok {
			invalid = append(invalid, p)
		}
	}

	for _, p := range invalid {
		delete(partitionConsumers, p)
	}

	return nil
}

func (demoConsumerGroupHandler) Cleanup(_ sarama.ConsumerGroupSession) error { return nil }

func (h demoConsumerGroupHandler) ConsumeClaim(sess sarama.ConsumerGroupSession, claim sarama.ConsumerGroupClaim) error {
	for msg := range claim.Messages() {
		if err := partitionConsumers[msg.Partition].consume(sess, msg); nil != err {
			return err
		}
	}
	return nil
}

type partitionMsgConsumer struct {
	lastPartitionSeq  uint64
	lastPartitionSeq2 uint64
	lastCommitTime    int64
}

func (pmc *partitionMsgConsumer) consume(sess sarama.ConsumerGroupSession, msg *sarama.ConsumerMessage) error {
	ps2, err := getPartitionSeq2(msg)
	// 忽略这里的逻辑，当前版本sync2kafka任务无需关注
	//var validPs uint64
	//if err == nil {
	//	if pmc.lastPartitionSeq2 > 0 {
	//		if ps2 < pmc.lastPartitionSeq2 {
	//			fmt.Printf("duplicated msg, partition:%v, offset:%v, maxSeq:%v, current:%v\n",
	//				msg.Partition, msg.Offset, pmc.lastPartitionSeq2, ps2)
	//			return nil
	//		} else if ps2 == pmc.lastPartitionSeq2 { // 可能是重复消息也可能是checkpoint或heartbeat消息
	//			if !isCheckpoint(msg) && !isHeartbeat(msg) {
	//				fmt.Printf("duplicated msg, partition:%v, offset:%v, maxSeq:%v, current:%v\n",
	//					msg.Partition, msg.Offset, pmc.lastPartitionSeq2, ps2)
	//				return nil
	//			}
	//		} else {
	//			if ps2 != pmc.lastPartitionSeq2+1 {
	//				msg := fmt.Sprintf("the PartitionSeq2 is discontinuous, last: %v, current: %v, "+
	//					"partition:%v, offset:%v", pmc.lastPartitionSeq2, ps2, msg.Partition, msg.Offset)
	//				panic(msg)
	//			}
	//		}
	//	}
	//	pmc.lastPartitionSeq2 = ps2
	//	validPs = ps2
	//} else {
	//	ps, err := getPartitionSeq(msg)
	//	if err != PsNotExistError {
	//		if ps <= pmc.lastPartitionSeq {
	//			fmt.Printf("duplicated msg: %+v\n", msg)
	//			return nil
	//		} else if ps != pmc.lastPartitionSeq+1 && pmc.lastPartitionSeq != 0 {
	//			msg := fmt.Sprintf("the PartitionSeq is discontinuous, last: %v, current: %v, "+
	//				"partition:%v, offset:%v", pmc.lastPartitionSeq, ps, msg.Partition, msg.Offset)
	//			panic(msg)
	//		}
	//		pmc.lastPartitionSeq = ps
	//		validPs = ps
	//	}
	//}

	record, err := avro.DeserializeRecord(msg.Value, false)
	if nil != err {
		// 如果反序列化失败，一般是因为任务是旧版本，需要使用旧版本的schema再次尝试
		// 注意：如果确定是旧版本任务，建议优化这里的逻辑，绕过第一次失败的尝试oldData直接传入true反序列化
		record, err = avro.DeserializeRecord(msg.Value, true)
		if nil != err {
			fmt.Printf("DeserializeRecord error: %v\n", err)
			panic(err.Error())
		}
	}

	err = onEntry(msg.Partition, msg.Offset, ps2, record)
	if err != nil {
		return err
	}

	/* 以下两种Kafka提交位点方式默认同时生效，如有特殊需要可以自行修改 */
	// commit type 1: 遇到checkpoint消息做一次提交
	if isCheckpoint(msg) {
		sess.MarkMessage(msg, "")
	}
	// commit type 2: 每隔10秒做一次提交
	if pmc.lastCommitTime == 0 || (time.Now().Unix()-pmc.lastCommitTime) >= 10 {
		sess.MarkMessage(msg, "")
		pmc.lastCommitTime = time.Now().Unix()
	}

	return nil
}

const partitionSeqKey = "ps"
const partitionSeq2Key = "ps2"

var PsNotExistError = errors.New(partitionSeqKey + " does not exists")

func getPartitionSeq(msg *sarama.ConsumerMessage) (uint64, error) {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeqKey {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq: " + string(m.Value))
			}
			return n, nil
		}
	}
	return 0, PsNotExistError
}

func getPartitionSeq2(msg *sarama.ConsumerMessage) (uint64, error) {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeq2Key {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq2: " + string(m.Value))
			}
			return n, nil
		}
	}
	return 0, errors.New(partitionSeq2Key + " does not exists")
}

func onEntry(partitionNum int32, offset int64, partitionSeq uint64, record *gen.Record) error {
	b, err := json.Marshal(record)
	if err != nil {
		return err
	}
	fmt.Printf("-->[partition: %v, offset: %v, partitionSeq: %v]\n %s\n", partitionNum, offset, partitionSeq,
		string(b))
	if *trans2sql {
		return trans2SQL(record)
	}
	return nil
}

func trans2SQL(record *gen.Record) error {
	switch record.MessageType {
	case gen.MessageTypeHEARTBEAT:
		fmt.Println("heartbeat:", record.HeartbeatTimestamp)
	case gen.MessageTypeCHECKPOINT:
		fmt.Println("Checkpoint: gtid=", record.SyncedGtid, ", file=", record.FileName, ", position=", record.Position)
	case gen.MessageTypeBEGIN:
		fmt.Println("BEGIN")
	case gen.MessageTypeCOMMIT:
		fmt.Println("COMMIT")
	case gen.MessageTypeROLLBACK:
		fmt.Println("ROLLBACK")
	case gen.MessageTypeDDL, gen.MessageTypeINIT_DDL, gen.MessageTypeSTATEMENT:
		fmt.Println(record.Sql)
	case gen.MessageTypeINSERT:
		if sql, err := transInsert(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	case gen.MessageTypeINIT_INSERT:
		if sql, err := transInitInsert(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	case gen.MessageTypeDELETE:
		if sql, err := transDelete(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	case gen.MessageTypeUPDATE:
		if sql, err := transUpdate(record); nil != err {
			return err
		} else {
			fmt.Println(sql)
		}
	default:
		return errors.New("unsupported message type")
	}

	return nil
}

func transInsert(record *gen.Record) (string, error) {
	switch record.NewColumns.UnionType {
	case avro.Array:
		if record.Columns != nil && record.Columns.UnionType == gen.UnionNullArrayFieldTypeEnumArrayField {
			sql := &strings.Builder{}
			sql.WriteString(fmt.Sprintf("INSERT INTO `%v`.`%v` VALUES ", record.Field_SchemaName, record.TableName))
			columns := record.Columns.ArrayField
			sql.WriteString("(")
			sep := ""
			for idx, value := range record.NewColumns.ArrayUnionNullIntegerCharacterDecimalFloatTimestampDateTimeTimestampWithTimeZoneBinaryGeometryTextGeometryBinaryObjectTextObjectEmptyObject {
				sql.WriteString(sep)

				if d, err := decode(columns[idx], value); nil != err {
					fmt.Println(err)
					return "", err
				} else {
					sql.WriteString(fmt.Sprintf("%v", d))
				}

				sep = ", "
			}
			sql.WriteString(")")
			return sql.String(), nil
		} else {
			return "", errors.New("columns definition is null")
		}
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %v", record.NewColumns.UnionType))
	}
}

func transInitInsert(record *gen.Record) (string, error) {
	switch record.NewColumns.UnionType {
	case avro.Array:
		if record.Columns != nil && record.Columns.UnionType == gen.UnionNullArrayFieldTypeEnumArrayField {
			sql := &strings.Builder{}
			cols := make([]string, 0, len(record.Columns.ArrayField))
			for _, col := range record.Columns.ArrayField {
				cols = append(cols, col.Name)
			}
			sql.WriteString(fmt.Sprintf("INSERT INTO `%v`.`%v`(%s) VALUES ", record.Field_SchemaName, record.TableName, strings.Join(cols, ",")))

			columns := record.Columns.ArrayField
			sql.WriteString("(")
			sep := ""
			for idx, value := range record.NewColumns.ArrayUnionNullIntegerCharacterDecimalFloatTimestampDateTimeTimestampWithTimeZoneBinaryGeometryTextGeometryBinaryObjectTextObjectEmptyObject {
				sql.WriteString(sep)

				if d, err := decode(columns[idx], value); nil != err {
					fmt.Println(err)
					return "", err
				} else {
					sql.WriteString(fmt.Sprintf("%v", d))
				}

				sep = ", "
			}
			sql.WriteString(")")
			return sql.String(), nil
		} else {
			return "", errors.New("columns definition is null")
		}
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %v", record.NewColumns.UnionType))
	}
}

func transDelete(record *gen.Record) (string, error) {
	switch record.OldColumns.UnionType {
	case avro.Array:
		if record.Columns != nil && record.Columns.UnionType == gen.UnionNullArrayFieldTypeEnumArrayField {
			sql := &strings.Builder{}
			sql.WriteString(fmt.Sprintf("DELETE FROM `%v`.`%v` WHERE ", record.Field_SchemaName, record.TableName))
			columns := record.Columns.ArrayField
			sep := ""
			for idx, value := range record.OldColumns.ArrayUnionNullIntegerCharacterDecimalFloatTimestampDateTimeTimestampWithTimeZoneBinaryGeometryTextGeometryBinaryObjectTextObjectEmptyObject {
				if value == nil {
					continue
				}

				sql.WriteString(sep)
				colDef := columns[idx]
				sql.WriteString(fmt.Sprintf("`%v`", colDef.Name))
				if value.UnionType == avro.EmptyObject {
					sql.WriteString(" IS NULL")
				} else if d, err := decode(colDef, value); nil != err {
					fmt.Println(err)
					return "", err
				} else {
					sql.WriteString(fmt.Sprintf(" = %v", d))
				}
				sep = " AND "
			}

			sql.WriteString(" LIMIT 1")
			return sql.String(), nil
		} else {
			return "", errors.New("columns definition is null")
		}
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %v", record.NewColumns.UnionType))
	}
}

func transUpdate(record *gen.Record) (string, error) {
	switch record.NewColumns.UnionType {
	case avro.Array:
		if record.Columns != nil && record.Columns.UnionType == gen.UnionNullArrayFieldTypeEnumArrayField {
			sql := &strings.Builder{}
			sql.WriteString(fmt.Sprintf("UPDATE `%v`.`%v` SET ", record.Field_SchemaName, record.TableName))
			columns := record.Columns.ArrayField
			sep := ""
			for idx, value := range record.NewColumns.ArrayUnionNullIntegerCharacterDecimalFloatTimestampDateTimeTimestampWithTimeZoneBinaryGeometryTextGeometryBinaryObjectTextObjectEmptyObject {
				if value == nil {
					continue
				}

				colDef := columns[idx]
				sql.WriteString(sep)
				sql.WriteString(fmt.Sprintf("`%v`", colDef.Name))
				if d, err := decode(colDef, value); nil != err {
					fmt.Println(err)
					return "", err
				} else {
					sql.WriteString(fmt.Sprintf(" = %v", d))
				}
				sep = ", "
			}

			sql.WriteString(" WHERE ")
			sep = ""
			for idx, value := range record.OldColumns.ArrayUnionNullIntegerCharacterDecimalFloatTimestampDateTimeTimestampWithTimeZoneBinaryGeometryTextGeometryBinaryObjectTextObjectEmptyObject {
				if value == nil {
					continue
				}

				sql.WriteString(sep)
				colDef := columns[idx]
				sql.WriteString(fmt.Sprintf("`%v`", colDef.Name))
				if value.UnionType == avro.EmptyObject {
					sql.WriteString(" IS NULL")
				} else if d, err := decode(colDef, value); nil != err {
					fmt.Println(err)
					return "", err
				} else {
					sql.WriteString(fmt.Sprintf(" = %v", d))
				}
				sep = " AND "
			}

			sql.WriteString(" LIMIT 1")
			return sql.String(), nil
		} else {
			return "", errors.New("columns definition is null")
		}
	default:
		// 其他类型暂不支持
		return "", errors.New(fmt.Sprintf("unsupported columns value type: %v", record.NewColumns.UnionType))
	}
}

func decode(field *gen.Field, object *gen.UnionNullIntegerCharacterDecimalFloatTimestampDateTimeTimestampWithTimeZoneBinaryGeometryTextGeometryBinaryObjectTextObjectEmptyObject) (string, error) {
	if object == nil {
		return "DEFAULT", nil
	}
	switch object.UnionType {
	case avro.Integer:
		return avro.ParseInteger(object.Integer), nil
	case avro.Character:
		v, err := avro.ParseCharacter(object.Character)
		if err != nil {
			fmt.Println(err)
			return "x'" + hex.EncodeToString(object.Character.Value) + "'", nil
		} else {
			return "_binary'" + v + "'", nil
		}
	case avro.Decimal:
		return avro.ParseDecimal(object.Decimal), nil
	case avro.Float:
		if strings.Contains(field.OriginalType, "float") {
			return avro.ParseFloat(object.Float), nil
		}
		return avro.ParseDouble(object.Float), nil
	case avro.Timestamp:
		return "_binary'" + avro.ParseTimestamp(object.Timestamp) + "'", nil
	case avro.DateTime:
		return "_binary'" + avro.ParseDateTime(object.DateTime, "",
			int(field.DataTypeNumber)) + "'", nil
	case avro.TimestampWithTimeZone:
		return "_binary'" + avro.ParseDateTime(object.TimestampWithTimeZone.Value,
			object.TimestampWithTimeZone.Timezone,
			int(field.DataTypeNumber)) + "'", nil
	case avro.BinaryGeometry:
		return "x'" + avro.ParseBinaryGeometry(object.BinaryGeometry) + "'", nil
	case avro.TextGeometry:
		return "_binary'" + avro.ParseTextGeometry(object.TextGeometry) + "'", nil
	case avro.BinaryObject:
		return "x'" + avro.ParseBinaryObject(object.BinaryObject) + "'", nil
	case avro.TextObject:
		if object.TextObject.Type == "json" {
			return "'" + avro.ParseTextObject(object.TextObject) + "'", nil
		}
		return "_binary'" + avro.ParseTextObject(object.TextObject) + "'", nil
	case avro.EmptyObject:
		return "NULL", nil
	default:
		return "", errors.New("unsupported data type")
	}
}

func isCheckpoint(msg *sarama.ConsumerMessage) bool {
	for _, m := range msg.Headers {
		if string(m.Key) == "Checkpoint" {
			return true
		}
	}
	return false
}

func isHeartbeat(msg *sarama.ConsumerMessage) bool {
	for _, m := range msg.Headers {
		if string(m.Key) == "Heartbeat" {
			return true
		}
	}
	return false
}
